"""
Unit-tests for the shortcuts module

Run as python test_shortcuts.py in console or via coverage
"""
from __future__ import annotations
import unittest
import numpy as np
import numpy.testing as nptest

import longitudinal_tomography.tracking.machine as mch
import longitudinal_tomography.shortcuts as shortcuts
from . import commons


MACHINE_ARGS = commons.get_machine_args()


class TestShortcuts(unittest.TestCase):
    def test_track(self):

        machine = mch.Machine(**MACHINE_ARGS)

        # Tracking only a few particles for 20 time frames.
        machine.snpt = 1
        rbn = 13
        machine.rbn = rbn
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.nprofiles = 20

        machine.values_at_turns()

        xp, yp = shortcuts.track(machine, 10)

        # Comparing the coordinates of particle #0 only.
        correct_x = [8, 8, 9, 10, 11, 5, 6, 7, 8, 9, 10,
                     11, 12, 12, 13, 14, 15, 4, 5, 6]

        correct_y = [14, 13, 13, 13, 12, 16, 16, 15, 15, 14,
                     14, 13, 13, 13, 12, 12, 11, 18, 18, 17]

        for x, cx in zip(xp[:, 0], correct_x):
            self.assertAlmostEqual(
                x, cx, msg='Error in tracking of particle '
                           'found in x-coordinate')
        for y, cy in zip(yp[:, 0], correct_y):
            self.assertAlmostEqual(
                y, cy, msg='Error in tracking of particle '
                           'found in y-coordinate')

    def test_tomogram(self):

        waterfall = commons.load_waterfall()

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0].T

        tomo = shortcuts.tomogram(waterfall, xp, yp=None, n_iter=1)

        weights = tomo.weight
        cweights = np.array([0.00861438, 0.00880599, 0.00893367, 0.00909539,
                             0.00918903, 0.00922729, 0.00928258, 0.00928246,
                             0.00927822, 0.00929515, 0.00929504, 0.00921845,
                             0.0092014,  0.00918014, 0.00915893, 0.00914618,
                             0.00919728, 0.00912925, 0.00918038, 0.00924009,
                             0.00920603, 0.00930828, 0.00940621, 0.00953821,
                             0.00959356, 0.00969566, 0.00981058, 0.00988708,
                             0.00998924, 0.01007859, 0.01014255, 0.01023618,
                             0.01021914, 0.01028299, 0.01026588, 0.01021909,
                             0.01016806, 0.01015526, 0.01011276, 0.01008299,
                             0.00999782, 0.00992973, 0.00991267, 0.00985308,
                             0.00977646, 0.0097126,  0.00971697, 0.00967434,
                             0.00961899, 0.00961908])

        nptest.assert_almost_equal(
            weights, cweights, err_msg='Weights were calculated incorrectly')



"""Module containing the reconstruction algorithm with CUDA kernels.

:Author(s): **Bernardo Abreu Figueiredo**
"""

import numpy as np
import cupy as cp
from ..utils.tomo_config import GPUDev
from ..utils.tomo_config import AppConfig as conf
import longitudinal_tomography.cuda_kernels as cuda_kernels

gpu_dev = GPUDev.get_gpu_dev()
block_size = gpu_dev.block_size

def refresh_kernels():
    global back_project_kernel, project_kernel, clip_kernel, find_diffprof_kernel,\
        count_part_bin_kernel, calc_reciprocal_kernel, comp_part_amount_kernel, create_flat_points_kernel

    gpu_dev = GPUDev.get_gpu_dev()
    back_project_kernel = gpu_dev.rec_mod.get_function("back_project")
    project_kernel = gpu_dev.rec_mod.get_function("project")
    clip_kernel = gpu_dev.rec_mod.get_function("clip")
    find_diffprof_kernel = gpu_dev.rec_mod.get_function("find_difference_profile")
    count_part_bin_kernel = gpu_dev.rec_mod.get_function("count_particles_in_bin")
    calc_reciprocal_kernel = gpu_dev.rec_mod.get_function("calculate_reciprocal")
    comp_part_amount_kernel = gpu_dev.rec_mod.get_function("compensate_particle_amount")
    create_flat_points_kernel = gpu_dev.rec_mod.get_function("create_flat_points")


def back_project(weights: cp.ndarray,
                 flat_points: cp.ndarray,
                 flat_profiles: cp.ndarray,
                 n_particles: int,
                 n_profiles: int) -> cp.ndarray:
    back_project_kernel(args=(weights, flat_points, flat_profiles, n_particles, n_profiles),
                            block=(cuda_kernels.REDUCTION_BLOCK_SIZE, 1, 1),
                            grid=(n_particles, 1, 1))
    return weights


def project(flat_rec: cp.ndarray,
            flat_points: cp.ndarray,
            weights: cp.ndarray, n_particles: int,
            n_profiles: int, n_bins: int) -> cp.ndarray:
    project_kernel(args=(flat_rec, flat_points, weights, n_particles, n_profiles),
                        block=block_size,
                        grid=(int((n_particles * n_profiles) / block_size[0] + 1), 1, 1))
    return flat_rec

def normalize(flat_rec: cp.ndarray,
              n_profiles: int, n_bins: int) -> cp.ndarray:
    flat_rec = flat_rec.reshape((n_profiles, n_bins))
    sum_profile = cp.sum(flat_rec, axis=1)
    flat_rec /= cp.expand_dims(sum_profile, axis=1)
    flat_rec = flat_rec.reshape(-1)
    sum_waterfall = cp.sum(sum_profile)

    if sum_waterfall <= 0:
        raise RuntimeError("Phase space reduced to zeros!")
    return flat_rec

def clip(array: cp.ndarray,
         array_length: int,
        clip_val: float) -> cp.ndarray:
    clip_kernel(args=(array, array_length, clip_val),
                block=block_size,
                grid=(int(array_length / block_size[0] + 1), 1, 1))
    return array

def find_difference_profile(flat_rec: cp.ndarray,
                            flat_profiles: cp.ndarray) -> cp.ndarray:
    length = len(flat_rec)
    diff_prof = cp.empty(length, dtype=flat_rec.dtype)
    find_diffprof_kernel(args=(diff_prof, flat_rec, flat_profiles, length),
                         block=block_size,
                         grid=(int(length / block_size[0] + 1), 1, 1))
    return diff_prof

def discrepancy(diff_prof: cp.ndarray,
                n_profiles: int, n_bins: int) -> float:
    all_bins = n_profiles * n_bins
    squared_sum = cp.sum(cp.power(diff_prof, 2))

    return cp.sqrt(squared_sum / all_bins)

def compensate_particle_amount(diff_prof: cp.ndarray,
                               rparts: cp.ndarray,
                               n_profiles: int, n_bins: int) -> cp.ndarray:
    comp_part_amount_kernel(args=(diff_prof, rparts, n_profiles, n_bins),
                            block=block_size,
                            grid=(int((n_profiles * n_bins) / block_size[0] + 1), 1, 1))
    return diff_prof

def max_2d(array: cp.ndarray,
           x_axis: int, y_axis: int) -> float:
    return cp.max(array[:y_axis, :x_axis])

def reciprocal_particles(rparts: cp.ndarray, xp: cp.ndarray,
                         n_bins: int, n_profiles: int,
                         n_particles: int) -> cp.ndarray:
    count_part_bin_kernel(args=(rparts, xp, n_profiles, n_particles, n_bins),
                          block=block_size,
                          grid=(int((n_particles * n_profiles) / block_size[0] + 1), 1, 1))

    max_bin_val = float(cp.max(rparts))

    calc_reciprocal_kernel(args=(rparts, n_bins, n_profiles, max_bin_val),
                           block=block_size,
                           grid=(int((n_bins * n_profiles) / block_size[0] + 1), 1, 1))

    return rparts

def create_flat_points(xp: cp.ndarray,
                       n_particles: int, n_profiles: int,
                       n_bins: int) -> cp.ndarray:
    flat_points = cp.copy(xp)

    create_flat_points_kernel(args=(flat_points, n_particles, n_profiles, n_bins),
                              block=block_size,
                              grid=(int((n_particles * n_profiles) / block_size[0] + 1), 1, 1))

    return flat_points

def reconstruct_cuda(xp: cp.ndarray,
                waterfall: cp.ndarray, n_iter: int,
                n_bins: int, n_particles: int, n_profiles: int,
                verbose: bool = False, callback = None) -> tuple:
    xp = xp.flatten()
    # from wrapper
    weights = cp.zeros(n_particles, dtype=conf.get_precision())
    discr = np.zeros(n_iter + 1)
    flat_profiles = waterfall.flatten().astype(conf.get_precision())
    flat_rec = cp.zeros(n_profiles * n_bins, dtype=conf.get_precision())
    flat_points = cp.zeros(n_particles * n_profiles)
    rparts = cp.zeros((n_profiles * n_bins), dtype=conf.get_precision())

    # Actual functionality
    rparts = reciprocal_particles(rparts, xp, n_bins, n_profiles, n_particles)
    flat_points = create_flat_points(xp, n_particles, n_profiles, n_bins)
    weights = back_project(weights, flat_points, flat_profiles, n_particles, n_profiles)
    weights = clip(weights, n_particles, 0.0)

    if cp.sum(weights) <= 0:
        raise RuntimeError("All of phase space got reduced to zeros")

    if verbose:
        print(" Iterating...")

    for iteration in range(n_iter):
        if verbose:
            print(f"{iteration+1:3}")

        flat_rec = project(flat_rec, flat_points, weights, n_particles, n_profiles, n_bins)
        flat_rec = normalize(flat_rec, n_profiles, n_bins)
        diff_prof = find_difference_profile(flat_rec, flat_profiles)
        discr[iteration] = discrepancy(diff_prof, n_profiles, n_bins)
        diff_prof = compensate_particle_amount(diff_prof, rparts, n_profiles, n_bins)
        weights = back_project(weights, flat_points, diff_prof, n_particles, n_profiles)
        weights = clip(weights, n_particles, 0.0)

        if cp.sum(weights) <= 0:
            raise RuntimeError("All of phase space got reduced to zeros")

    # Calculating final discrepancy
    flat_rec = project(flat_rec, flat_points, weights, n_particles, n_profiles, n_bins)
    flat_rec = normalize(flat_rec, n_profiles, n_bins)
    diff_prof = find_difference_profile(flat_rec, flat_profiles)
    discr[n_iter] = discrepancy(diff_prof, n_profiles, n_bins)

    if verbose:
        print("Done!")

    return weights, discr, flat_rec

refresh_kernels()


from __future__ import annotations
import re
import time
import typing as t
from copy import deepcopy
from os import path

import numpy as np
import yaml

import longitudinal_tomography.shortcuts as shortcuts
import longitudinal_tomography.utils.tomo_input as tomoin
from longitudinal_tomography.data.profiles import Profiles
from longitudinal_tomography.tomography import Tomography
from longitudinal_tomography.tracking import Machine
from longitudinal_tomography.utils.tomo_input import Frames

_tomo: Tomography = None
_processing_tomo = False
_waterfall: np.ndarray = None
_machine: Machine = None
_frames: Frames = None
_profiles: Profiles = None

test_root = path.split(path.abspath(__file__))[0]


def load_data() -> t.Tuple[Machine, Frames, Profiles]:
    global _machine, _frames, _profiles
    if _machine is not None:
        return _machine, _frames, _profiles

    base_dir = path.split(path.realpath(__file__))[0]
    data_path = path.join(base_dir, 'resources')
    dat_path = path.join(data_path, 'INDIVShavingC325.dat')
    raw_params, raw_data = tomoin.get_user_input(dat_path)

    machine, frames = tomoin.txt_input_to_machine(raw_params)
    machine.values_at_turns()
    waterfall = frames.to_waterfall(raw_data)

    profiles = tomoin.raw_data_to_profiles(
        waterfall, machine, frames.rebin, frames.sampling_time)
    profiles.calc_profilecharge()

    _machine = machine
    _frames = frames
    _profiles = profiles

    return machine, frames, profiles


def get_tomography_params() -> t.Tuple[Tomography, Machine]:
    global _tomo, _processing_tomo
    if _tomo is not None or _processing_tomo:
        while _processing_tomo:
            time.sleep(0.1)
        return deepcopy(_tomo), deepcopy(_machine)
    if _machine is None:
        load_data()

    _processing_tomo = True
    xp, yp = shortcuts.track(_machine, 0)
    tomo = shortcuts.tomogram(_profiles.waterfall, xp, yp, 2)

    _tomo = tomo
    _processing_tomo = False
    return deepcopy(tomo), deepcopy(_machine)


def get_machine_args() -> dict:
    test_root = path.split(path.abspath(__file__))[0]
    yml_path = path.join(test_root, 'machine_args.yml')

    with open(yml_path, 'r') as f:
        machine_args = yaml.full_load(f)

    # parse mathematical expressions
    regex = re.compile(r'.+[\*\+\-\/].+')
    for k, v in machine_args.items():
        if isinstance(v, str) and regex.search(v):
            possible_v = eval(v)
            print(type(possible_v))
            if isinstance(possible_v, int) \
                    or isinstance(possible_v, float):
                machine_args[k] = possible_v

    return machine_args


def load_waterfall():
    base_dir = path.split(path.realpath(__file__))[0]
    data_path = path.join(base_dir, 'resources')

    waterfall = np.load(path.join(
        data_path, 'waterfall_INDIVShavingC325.npy'))
    return waterfall



"""Unit-tests for the functions in the particles module.

Run as python test_particles_class.py in console or via coverage
"""
from __future__ import annotations
import unittest

import numpy as np

from .. import commons
import longitudinal_tomography.tracking.machine as mch
import longitudinal_tomography.tracking.particles as pts
from longitudinal_tomography.utils import tomo_config as conf
from longitudinal_tomography import exceptions as expt

# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestParticlesMethods(unittest.TestCase):

    def test_filter_lost_correct(self):
        xp = np.array([[1, 2, 3, 4, 5, 6, 7],
                       [1, 2, 3, 4, 5, 6, 7],
                       [1, 2, 3, 3, 3, 4, 4],
                       [1, 2, 3, 3, 3, 4, 6],
                       [1, 2, 3, 3, 3, 4, 1],
                       [1, 2, 5, 3, 3, 4, 1]])
        yp = np.ones(xp.shape)

        img_width = 5
        xp, yp, nr_lost = pts.filter_lost(xp, yp, img_width)

        correct_xp = np.array([[1, 2, 4], [1, 2, 4], [1, 2, 3],
                               [1, 2, 3], [1, 2, 3], [1, 2, 3]])
        correct_yp = np.ones((6, 3))

        for xvec, cxvec in zip(xp, correct_xp):
            for x, cx in zip(xvec, cxvec):
                self.assertEqual(x, cx, msg='Error in xp coordinates '
                                            'after filtering')

        for yvec, cyvec in zip(yp, correct_yp):
            for y, cy in zip(yvec, cyvec):
                self.assertEqual(y, cy, msg='Error in yp coordinates '
                                            'after filtering')

        self.assertEqual(nr_lost, 4, msg='Error in reported number of '
                                         'lost particles')

    def test_filter_lost_all_pts_lost_error(self):
        xp = np.array([[5, 6, 7], [5, 6, 7], [3, 4, 4],
                       [3, 4, 6], [3, 4, 1], [3, 4, 1]])
        yp = np.ones(xp.shape)

        img_width = 5

        with self.assertRaises(expt.InvalidParticleError,
                               msg='Removing all particles should '
                                   'raise an exception'):
            xp, yp, nr_lost = pts.filter_lost(xp, yp, img_width)

    def test_physical_to_coords_correct(self):
        phases = [[-0.24180582, 0.04498875, 0.33178332, 0.61857789],
                  [-0.24180582, 0.04498875, 0.33178332, 0.61857789]]
        energies = [[-115567.32591061, -115567.32591061, -115567.32591061,
                     -115567.32591061],
                    [-38522.4419702,   -38522.4419702, -38522.4419702,
                     -38522.4419702]]

        phases = np.array(phases)
        energies = np.array(energies)

        machine = mch.Machine(**MACHINE_ARGS)

        machine.dturns = 5
        machine.phi0 = np.array([0.40078213, 0.40078213, 0.40078213,
                                 0.40078213, 0.40078213, 0.40078213])
        machine.h_num = 1
        machine.omega_rev0 = np.array([4585866.32214847, 4585893.44719289,
                                       4585920.57199441, 4585947.69655305,
                                       4585974.82086881, 4586001.94494169])
        machine.dtbin = 9.999999999999999e-10
        machine.synch_part_y = 380.0

        xorigin = -246.60492626420734
        dEbin = 1232.7181430465346

        xp, yp = pts.physical_to_coords(
            phases, energies, machine, xorigin, dEbin)

        correct_xp = [[281.27150807, 343.8103066,  406.34910513, 468.88790365],
                      [281.27048287, 343.80743192, 406.34438098, 468.88133003]]
        correct_yp = [[286.25, 286.25, 286.25, 286.25],
                      [348.75, 348.75, 348.75, 348.75]]

        for xvec, cxvec in zip(xp, correct_xp):
            for x, cx in zip(xvec, cxvec):
                self.assertAlmostEqual(
                    x, cx, msg='Error in calculated xp coordinates ')

        for yvec, cyvec in zip(yp, correct_yp):
            for y, cy in zip(yvec, cyvec):
                self.assertAlmostEqual(
                    y, cy, msg='Error in calculated yp coordinates ')

    def test_physical_to_coords_error(self):
        phases = [[-0.24180582, 0.04498875, 0.33178332, 0.61857789],
                  [-0.24180582, 0.04498875, 0.33178332, 0.61857789]]
        energies = [[-115567.32591061, -115567.32591061, -115567.32591061],
                    [-38522.4419702, -38522.4419702, -38522.4419702]]

        phases = np.array(phases)
        energies = np.array(energies)

        machine = mch.Machine(**MACHINE_ARGS)

        machine.dturns = 5
        machine.phi0 = np.array([0.40078213, 0.40078213, 0.40078213,
                                 0.40078213, 0.40078213, 0.40078213])
        machine.h_num = 1
        machine.omega_rev0 = np.array([4585866.32214847, 4585893.44719289,
                                       4585920.57199441, 4585947.69655305,
                                       4585974.82086881, 4586001.94494169])
        machine.dtbin = 9.999999999999999e-10
        machine.synch_part_y = 380.0

        xorigin = -246.60492626420734
        dEbin = 1232.7181430465346

        with self.assertRaises(expt.InvalidParticleError,
                               msg='Providing arrays of coordinates of '
                                   'different lengths should raise '
                                   'an exception'):
            xp, yp = pts.physical_to_coords(
                phases, energies, machine, xorigin, dEbin)

class TestParticlesMethodsGPU(unittest.TestCase):

    def setUp(self):
        try:
            import cupy as cp
            self.cp = cp
            conf.AppConfig.use_gpu()
            conf.AppConfig.set_double_precision()
        except ImportError:
            self.skipTest('CuPy not found - skipped tests')

    def tearDown(self) -> None:
        conf.AppConfig.use_cpu()
        conf.AppConfig.set_double_precision()

    def test_filter_lost_correct(self):
        xp = self.cp.array([[1, 2, 3, 4, 5, 6, 7],
                            [1, 2, 3, 4, 5, 6, 7],
                            [1, 2, 3, 3, 3, 4, 4],
                            [1, 2, 3, 3, 3, 4, 6],
                            [1, 2, 3, 3, 3, 4, 1],
                            [1, 2, 5, 3, 3, 4, 1]])
        yp = self.cp.ones(xp.shape)

        img_width = 5
        xp, yp, nr_lost = pts.filter_lost(xp, yp, img_width)

        correct_xp = self.cp.array([[1, 2, 4], [1, 2, 4], [1, 2, 3],
                                    [1, 2, 3], [1, 2, 3], [1, 2, 3]])
        correct_yp = self.cp.ones((6, 3))

        for xvec, cxvec in zip(xp, correct_xp):
            for x, cx in zip(xvec, cxvec):
                self.assertEqual(x, cx, msg='Error in xp coordinates '
                                            'after filtering')

        for yvec, cyvec in zip(yp, correct_yp):
            for y, cy in zip(yvec, cyvec):
                self.assertEqual(y, cy, msg='Error in yp coordinates '
                                            'after filtering')

        self.assertEqual(nr_lost, 4, msg='Error in reported number of '
                                         'lost particles')

    def test_filter_lost_all_pts_lost_error(self):
        xp = self.cp.array([[5, 6, 7], [5, 6, 7], [3, 4, 4],
                            [3, 4, 6], [3, 4, 1], [3, 4, 1]])
        yp = self.cp.ones(xp.shape)

        img_width = 5

        with self.assertRaises(expt.InvalidParticleError,
                               msg='Removing all particles should '
                                   'raise an exception'):
            xp, yp, nr_lost = pts.filter_lost(xp, yp, img_width)

    def test_physical_to_coords_correct(self):
        phases = [[-0.24180582, 0.04498875, 0.33178332, 0.61857789],
                  [-0.24180582, 0.04498875, 0.33178332, 0.61857789]]
        energies = [[-115567.32591061, -115567.32591061, -115567.32591061,
                     -115567.32591061],
                    [-38522.4419702,   -38522.4419702, -38522.4419702,
                     -38522.4419702]]

        phases = self.cp.array(phases)
        energies = self.cp.array(energies)

        machine = mch.Machine(**MACHINE_ARGS)

        machine.dturns = 5
        machine.phi0 = np.array([0.40078213, 0.40078213, 0.40078213,
                                 0.40078213, 0.40078213, 0.40078213])
        machine.h_num = 1
        machine.omega_rev0 = np.array([4585866.32214847, 4585893.44719289,
                                       4585920.57199441, 4585947.69655305,
                                       4585974.82086881, 4586001.94494169])
        machine.dtbin = 9.999999999999999e-10
        machine.synch_part_y = 380.0

        xorigin = -246.60492626420734
        dEbin = 1232.7181430465346

        xp, yp = pts.physical_to_coords(
            phases, energies, machine, xorigin, dEbin)

        correct_xp = [[281.27150807, 343.8103066,  406.34910513, 468.88790365],
                      [281.27048287, 343.80743192, 406.34438098, 468.88133003]]
        correct_yp = [[286.25, 286.25, 286.25, 286.25],
                      [348.75, 348.75, 348.75, 348.75]]

        for xvec, cxvec in zip(xp, correct_xp):
            for x, cx in zip(xvec, cxvec):
                self.assertAlmostEqual(
                    x.item(), cx, msg='Error in calculated xp coordinates ')

        for yvec, cyvec in zip(yp, correct_yp):
            for y, cy in zip(yvec, cyvec):
                self.assertAlmostEqual(
                    y.item(), cy, msg='Error in calculated yp coordinates ')

    def test_physical_to_coords_error(self):
        phases = self.cp.array([[-0.24180582, 0.04498875, 0.33178332, 0.61857789],
                                [-0.24180582, 0.04498875, 0.33178332, 0.61857789]])
        energies = self.cp.array([[-115567.32591061, -115567.32591061, -115567.32591061],
                                  [-38522.4419702, -38522.4419702, -38522.4419702]])

        phases = self.cp.array(phases)
        energies = self.cp.array(energies)

        machine = mch.Machine(**MACHINE_ARGS)

        machine.dturns = 5
        machine.phi0 = np.array([0.40078213, 0.40078213, 0.40078213,
                                 0.40078213, 0.40078213, 0.40078213])
        machine.h_num = 1
        machine.omega_rev0 = np.array([4585866.32214847, 4585893.44719289,
                                       4585920.57199441, 4585947.69655305,
                                       4585974.82086881, 4586001.94494169])
        machine.dtbin = 9.999999999999999e-10
        machine.synch_part_y = 380.0

        xorigin = -246.60492626420734
        dEbin = 1232.7181430465346

        with self.assertRaises(expt.InvalidParticleError,
                               msg='Providing arrays of coordinates of '
                                   'different lengths should raise '
                                   'an exception'):
            xp, yp = pts.physical_to_coords(
                phases, energies, machine, xorigin, dEbin)



"""Module containing the ProgramsMachine class

:Author(s): **Anton Lu**
"""
from __future__ import annotations
import numpy as np
import scipy.constants as c
import logging
import typing as t
from scipy import optimize
from scipy import constants as cont

from ..utils import physics
from .machine_base import MachineABC
from .. import assertions as asrt
from .. import exceptions as ex

log = logging.getLogger(__name__)


class ProgramsMachine(MachineABC):
    """
    An alternative variant of the Machine class that is intended to be used
    with precomputed voltage, momentum, and phase programs.

    The voltage, momentum and phases are interpolated at each turn from the
    passed numpy arrays as opposed to linearly extrapolated or considered
    constant as in the :py:tomo.tracking.Machine:

    See superclass for documentation about inherited class variables.

    Parameters
    ----------
    voltage_function: np.ndarray
        A 2d array of shape (n_harmonics, :) containing the precomputed voltage
        program. The first row of the array must be the time stamps of the
        voltages in the second and third rows which contains the voltage
        amplitudes. 1 or 2 harmonics are supported.
    phase_program: np.ndarray
        A 2d array of shape (n_harmonics, :) containing the precomputed phase
        program. The first row of the array must be the time stamps of the
        phases in the second and third rows which contains the phase
        shifts. 1 or 2 harmonics are supported.
    momentum_program: np.ndarray
        A 2d array of shape (2, :) containing the precomputed momentum
        program. The first row of the array must be the time stamps of the
        momentum in the second row which contains the momentum values.
    harmonics: list
        A list of integers where each is the harmonic of an RF station.
        The number of harmonics must be the same as the passed voltage and
        phase functions.
    t_ref: float
        The cTime at which to the machine parameters are referenced.
        Must withing the time range defined in the voltage, phase and
        momentum programs.
    vat_now: bool
        Calculate values at turns immediately after initialisation.

    """

    def __init__(self,
                 dturns: int,
                 voltage_function: np.ndarray,
                 momentum_function: np.ndarray,
                 harmonics: t.List[int],
                 mean_orbit_rad: float, bending_rad: float,
                 trans_gamma: float, rest_energy: float,
                 nprofiles: int,
                 nbins: int,
                 dtbin: float,
                 t_ref: float,
                 phase_function: np.ndarray = None,
                 vat_now: bool = True,
                 **kwargs):
        asrt.assert_inrange(len(harmonics), 'harmonics', 1, 2,
                            ex.ArrayLengthError,
                            'Only 1 or 2 harmonics accepted.')
        super().__init__(dturns, mean_orbit_rad, bending_rad, trans_gamma,
                         rest_energy, nprofiles, nbins, dtbin, **kwargs)

        self.voltage_raw = voltage_function
        self.momentum_raw = momentum_function
        self.harmonics = harmonics

        if 'phi12' in kwargs:
            if phase_function is not None:
                raise ValueError('You can only pass either phi12 or '
                                 'phase_function, not both.')
            self.phase_raw = None
            self.process_phase = 0
        elif phase_function is not None:
            self.phase_raw = phase_function
            self.process_phase = 1
        else:
            log.debug('Neither phi12 nor phase program passed. Must compute '
                      'phi12 separately.')
            self.process_phase = 2

        self.circumference: float = 2 * np.pi * self.mean_orbit_rad
        self.t_ref = t_ref

        # init variables
        self.momentum_function: np.ndarray = None
        self.bdot: np.ndarray = None
        self.n_turns: int = None
        self.i0: int = None

        if vat_now:
            self.values_at_turns()

    def values_at_turns(self):
        """
        Calculated function values at each turn. The discrete momentum, voltage
        and phase programs are interpolated to each turn.
        """
        momentum_time, momentum_function = self._interpolate_momentum(
            self.momentum_raw[0, :],
            self.momentum_raw[1, :]
        )

        # interpolate first harmonic voltage & phase
        self.vrf1_at_turn = np.interp(
            momentum_time,  # same time steps as momentum program
            self.voltage_raw[0, :],  # voltage time
            self.voltage_raw[1, :]  # voltage values
        )

        if self.process_phase == 1:
            phase1 = np.interp(
                momentum_time,
                self.phase_raw[0, :],
                self.phase_raw[1, :]
            )
        elif self.process_phase == 0:
            if isinstance(self.phi12, np.ndarray):
                if self.phi12.size() == 1:
                    pass
                elif len(self.phi12.shape) == 2 and self.phi12.shape[1] == 2:
                    self.phi12 = np.interp(momentum_time, *self.phi12)
                elif len(self.phi12) == len(self.voltage_raw):
                    self.phi12 = np.interp(momentum_time, self.voltage_raw[0],
                                           self.phi12)
                else:
                    raise ValueError(f'Do not know how to interpolate phi12 '
                                     f'array of shape {self.phi12.shape}. '
                                     f'It should either match voltage '
                                     f'function or provide its own time axis.')

        # interpolate second harmonic voltage & phase if applicable
        if len(self.harmonics) == 2:
            self.vrf2_at_turn = np.interp(
                momentum_time,
                self.voltage_raw[0, :],
                self.voltage_raw[2, :]
            )
            self.h_ratio = self.harmonics[1] / self.harmonics[0]

            if self.process_phase == 1:
                phase2 = np.interp(momentum_time,
                                   self.phase_raw[0, :],
                                   self.phase_raw[2, :])
        else:
            self.vrf2_at_turn = np.zeros_like(self.vrf1_at_turn)
            phase2 = 0
            self.h_ratio = 1

        self.n_turns = len(momentum_time)

        momentum = momentum_function
        energy = np.sqrt(momentum ** 2 + self.e_rest ** 2)
        deltaE0 = np.diff(energy)
        gamma = np.sqrt(1 + (momentum / self.e_rest) ** 2)

        beta0 = np.sqrt(1 / (1 + (self.e_rest / momentum) ** 2))
        t_rev = np.dot(self.circumference, 1 / (beta0 * c.c))
        f_rev = 1 / t_rev
        time_at_turn = np.cumsum(t_rev)
        time_at_turn -= time_at_turn[self.i0]

        omega_rev0 = 2 * np.pi * f_rev
        if self.process_phase == 1:
            self.phi12 = (phase1 - phase2 + np.pi) / self.h_ratio
        elif self.process_phase == 2:
            self.phi12 = None

        eta0 = (1. - beta0 ** 2) - self.trans_gamma ** (-2)
        drift_coef = (2 * np.pi * self.h_num * eta0
                      / (energy * beta0 ** 2))

        bfield = momentum / (self.bending_rad * self.q) / cont.c
        bdot = np.gradient(bfield, time_at_turn)

        self.momentum_function = momentum_function

        self.bdot = bdot
        self.beta0 = beta0
        self.deltaE0 = deltaE0
        self.drift_coef = drift_coef
        self.e0 = energy
        self.eta0 = eta0
        self.omega_rev0 = omega_rev0
        self.phi0 = np.zeros_like(self.vrf1_at_turn)
        self.time_at_turn = time_at_turn

        if self.process_phase != 2:
            self.find_synchronous_phase()

    def calc_phi12(self, c04mode: int):
        """
        Calculates the phase difference between the two RF stations as seen
        by the particle.

        N.B. Do not forget to run find_synchronous_phase after this!

        Parameters
        ----------
        c04mode: int
            0, 1, or 2, representing
            * Bunch shortening mode
            * Bunch lengthening mode
            * h2 + 1 mode
            respectively
        """
        log.info(f'Calculating phi12 and overwriting previously calculated '
                 f'values.')

        self.phi12 = np.arcsin(2 * np.pi * self.bending_rad *
                               self.mean_orbit_rad *
                               self.bdot / self.vrf1_at_turn) \
            + np.pi / self.h_ratio

        if c04mode == 2:
            self.phi12 -= np.pi

    def find_synchronous_phase(self):

        if self.phi12 is None:
            raise ValueError('phi12 is None. You must pass a phase program, '
                             'phi12, or call the calc_phi12 method to '
                             'calculate phi12.')

        phi_lower, phi_upper = physics.find_phi_lower_upper(self, self.i0)

        self.phi0 = np.zeros_like(self.vrf1_at_turn)

        try:
            phi_start = optimize.newton(func=physics.rfvolt_rf1_pmch,
                                        x0=(phi_lower + phi_upper) / 2.0,
                                        fprime=physics.drfvolt_rf1_pmch,
                                        tol=0.0001,
                                        maxiter=100,
                                        args=(self, self.i0))
            synch_phase = optimize.newton(func=physics.rf_voltage_pmch,
                                          x0=phi_start,
                                          fprime=physics.drf_voltage_pmch,
                                          tol=0.0001,
                                          maxiter=100,
                                          args=(self, self.i0))
            self.phi0[self.i0] = synch_phase

            for i in range(self.i0 + 1, self.n_turns):
                self.phi0[i] = optimize.newton(func=physics.rf_voltage_pmch,
                                               x0=self.phi0[i - 1],
                                               fprime=physics.drf_voltage_pmch,
                                               tol=0.0001,
                                               maxiter=100,
                                               args=(self, i))
            for i in range(self.i0 - 1, -1, -1):
                self.phi0[i] = optimize.newton(func=physics.rf_voltage_pmch,
                                               x0=self.phi0[i + 1],
                                               fprime=physics.drf_voltage_pmch,
                                               tol=0.0001,
                                               maxiter=100,
                                               args=(self, i))
        except RuntimeError:
            raise ValueError('Could not find synchronous phase for supplied '
                             'parameters.')

    def _interpolate_momentum(self, time: np.ndarray, momentum: np.ndarray):
        """
        Interpolates momentum at each turn. The interpolated time array is
        then used to interpolating voltage and phase programs.

        The interpolation code is borrowed from CERN BLonD.

        Parameters
        ----------
        time: np.ndarray
            A 1d array with the time codes for each momentum value.
        momentum: np.ndarray
            A 1d array with momentum values, each corresponding to a timestamp.

        Returns
        -------
        tuple(np.ndarray, np.ndarray)
            Interpolated time and momentum at each turn.
        """

        beta_0 = np.sqrt(1 / (1 + (self.e_rest / momentum[0]) ** 2))
        T0 = self.circumference / (
                beta_0 * c.c)  # Initial revolution period [s]
        time_interp = [time[0] + T0]
        beta_interp = [beta_0]
        momentum_interp = [momentum[0]]

        # Interpolate data recursively
        time_interp.append(time_interp[-1]
                           + self.circumference / (beta_interp[0] * c.c))

        nturns = self.dturns * self.nprofiles
        self.i0 = self.machine_ref_frame * self.dturns
        i = 0
        turn = self.i0

        k = 0
        while turn < nturns:

            while time_interp[i + 1] <= time[k]:
                momentum_interp.append(
                    momentum[k - 1] + (momentum[k] - momentum[k - 1]) *
                    (time_interp[i + 1] - time[k - 1]) /
                    (time[k] - time[k - 1]))

                beta_interp.append(
                    np.sqrt(
                        1 / (1 + (self.e_rest / momentum_interp[i + 1]) ** 2)))

                time_interp.append(
                    time_interp[i + 1] + self.circumference / (
                            beta_interp[i + 1] * c.c))

                i += 1

                if time_interp[i + 1] > self.t_ref:
                    turn += 1

            k += 1

        time_interp.pop()
        time_interp = np.asarray(time_interp)
        beta_interp = np.asarray(beta_interp)
        momentum_interp = np.asarray(momentum_interp)

        # Cutting the input momentum on the desired cycle time
        if self.t_ref is not None:
            initial_index = np.min(np.where(time_interp >= self.t_ref)[0])
            initial_index -= self.i0
        else:
            initial_index = 0

        final_index = initial_index + nturns
        if final_index > len(time_interp) + 1:
            raise ValueError('Not enough data in momentum program to '
                             f'interpolate {nturns} turns.')

        momentum_time = time_interp[initial_index:final_index]
        momentum_function = momentum_interp[initial_index:final_index]

        return momentum_time, momentum_function



"""Unit-tests for the physics module.

Run as python test_data_treatment.py in console or via coverage
"""
from __future__ import annotations
import os
import unittest

import numpy as np
import numpy.testing as nptest

from .. import commons
import longitudinal_tomography.data.data_treatment as treat
import longitudinal_tomography.data.profiles as prf
import longitudinal_tomography.tracking.machine as mch

# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestDataTreatment(unittest.TestCase):
    pass



"""Module containing the kick-and-drift algorithm with CUDA kernels.

:Author(s): **Bernardo Abreu Figueiredo**
"""

import numpy as np
import cupy as cp
import logging
from typing import Tuple
from ..utils.tomo_config import GPUDev

log = logging.getLogger(__name__)

gpu_dev = GPUDev.get_gpu_dev()
block_size = gpu_dev.block_size

def refresh_kernels():
    global kick_drift_up_turns, kick_drift_down_turns
    gpu_dev = GPUDev.get_gpu_dev()
    kick_drift_up_turns = gpu_dev.kd_mod.get_function("kick_drift_up_turns")
    kick_drift_down_turns = gpu_dev.kd_mod.get_function("kick_drift_down_turns")

def kick_drift_up_whole(dphi: cp.ndarray, denergy: cp.ndarray, xp: cp.ndarray, yp: cp.ndarray, drift_coef: cp.ndarray,
                        rfv1: cp.ndarray, rfv2: cp.ndarray, phi0: cp.ndarray, phi12: cp.ndarray, h_ratio: float,
                        n_particles: int, acc_kick: cp.ndarray, turn: int, nturns: int, dturns: int, profile: int) -> None:
    kick_drift_up_turns(args=(dphi, denergy, xp, yp, drift_coef, rfv1, rfv2,
                            phi0, phi12, h_ratio, n_particles, acc_kick,
                            turn, nturns, dturns, profile),
                        block=block_size, grid=(int(n_particles / block_size[0] + 1), 1, 1))

def kick_drift_down_whole(dphi: cp.ndarray, denergy: cp.ndarray, xp: cp.ndarray, yp: cp.ndarray, drift_coef: cp.ndarray,
                        rfv1: cp.ndarray, rfv2: cp.ndarray, phi0: cp.ndarray, phi12: cp.ndarray, h_ratio: float,
                        n_particles: int, acc_kick: cp.ndarray, turn: int, dturns: int, profile: int) -> None:
    kick_drift_down_turns(args=(dphi, denergy, xp, yp, drift_coef, rfv1, rfv2,
                            phi0, phi12, h_ratio, n_particles, acc_kick,
                            turn, dturns, profile),
                        block=block_size, grid=(int(n_particles / block_size[0] + 1), 1, 1))

def kick_and_drift_cuda(xp: cp.ndarray, yp: cp.ndarray,
                   denergy: cp.ndarray, dphi: cp.ndarray,
                   rfv1: np.ndarray, rfv2: np.ndarray,
                   phi0: np.ndarray,
                   deltaE0: np.ndarray,
                   drift_coef: np.ndarray,
                   phi12: float,
                   h_ratio: float,
                   dturns: int,
                   rec_prof: int,
                   deltaturn: int,
                   nturns: int,
                   nparts: int,
                   ftn_out: bool,
                   callback
                   ) -> Tuple[cp.ndarray, cp.ndarray]:
    phi12_arr = cp.full(nturns+1, phi12)
    # Preparation end

    profile = rec_prof
    turn = rec_prof * dturns + deltaturn

    if deltaturn < 0:
        profile -= 1

    # Value-based copy to avoid side-effects
    xp[profile] = cp.copy(dphi)
    yp[profile] = cp.copy(denergy)

    kick_drift_up_whole(dphi, denergy, xp, yp, drift_coef, rfv1, rfv2,
                        phi0, phi12_arr, h_ratio, nparts, deltaE0,
                        turn, nturns, dturns, profile)

    profile = rec_prof
    turn = rec_prof * dturns


    if profile > 0:
        # going back to initial coordinates
        dphi = cp.copy(xp[rec_prof])
        denergy = cp.copy(yp[rec_prof])

        kick_drift_down_whole(dphi, denergy, xp, yp, drift_coef, rfv1, rfv2,
                        phi0, phi12_arr, h_ratio, nparts, deltaE0,
                        turn, dturns, profile)

refresh_kernels()


"""Unit-tests for the functions in the tomolib_wrappers module.

Run as python test_tomolib_wrappers.py in console or via coverage
"""
from __future__ import annotations
import os
import unittest

import numpy as np
import numpy.testing as nptest

from .. import commons
import longitudinal_tomography.tracking.machine as mch
from longitudinal_tomography.cpp_routines import libtomo

# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestTLW(unittest.TestCase):

    def test_kick_up_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = 10
        machine.vrf2 = 500
        machine.values_at_turns()

        denergy = np.array([-115567.32591061])
        dphi = np.array([0.33178332])
        npart = 1
        turn = 0

        rfv1 = machine.vrf1_at_turn * machine.q
        rfv2 = machine.vrf2_at_turn * machine.q

        new_denergy = libtomo.kick(machine, denergy, dphi,
                                   rfv1, rfv2, npart, turn, up=True)

        correct_energy = -113495.65825924404
        self.assertAlmostEqual(
            new_denergy[0], correct_energy,
            msg='Kick in upward direction was calculated incorrectly')

    def test_kick_down_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = 10
        machine.vrf2 = 500
        machine.values_at_turns()

        denergy = np.array([-115567.32591061])
        dphi = np.array([0.33178332])
        npart = 1
        turn = (machine.nprofiles - 1) * machine.dturns

        rfv1 = machine.vrf1_at_turn * machine.q
        rfv2 = machine.vrf2_at_turn * machine.q

        new_denergy = libtomo.kick(machine, denergy, dphi,
                                   rfv1, rfv2, npart, turn, False)

        correct_energy = -116610.12118255378
        self.assertAlmostEqual(
            new_denergy[0], correct_energy,
            msg='Kick in downward direction was calculated incorrectly')

    def test_drift_up_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = 10
        machine.values_at_turns()

        denergy = np.array([-115567.32591061])
        dphi = np.array([0.33178332])
        nparts = 1
        turn = 0

        new_dphi = libtomo.drift(denergy, dphi, machine.drift_coef,
                                 nparts, turn, up=True)

        correct_dphi = 0.3356669466375665
        self.assertAlmostEqual(
            new_dphi[0], correct_dphi,
            msg='Drift in upward direction was calculated incorrectly')

    def test_drift_down_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = 10
        machine.values_at_turns()

        denergy = np.array([-115567.32591061])
        dphi = np.array([0.33178332])
        nparts = 1
        turn = (machine.nprofiles - 1) * machine.dturns

        new_dphi = libtomo.drift(denergy, dphi, machine.drift_coef,
                                 nparts, turn, up=False)

        correct_dphi = 0.3279023169434031
        self.assertAlmostEqual(
            new_dphi[0], correct_dphi,
            msg='Drift in downward direction was calculated incorrectly')

    def test_kick_and_drift_machine_arg_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = 10
        machine.values_at_turns()

        denergy = np.array([-115567.32591061])
        dphi = np.array([0.33178332])
        nparts = 1

        recprof = 5
        nturns = machine.dturns * (machine.nprofiles - 1)

        rfv1 = machine.vrf1_at_turn * machine.q
        rfv2 = machine.vrf2_at_turn * machine.q

        xp = np.zeros((machine.nprofiles, nparts))
        yp = np.zeros((machine.nprofiles, nparts))

        xp, yp = libtomo.kick_and_drift(
            xp, yp, denergy, dphi, rfv1, rfv2, machine, recprof, 0,
            nturns, nparts)

        correct_xp = np.array([[0.22739336], [0.24930078], [0.27073013],
                               [0.291644], [0.31200651], [0.33178332],
                               [0.35094168], [0.36945041], [0.38727993],
                               [0.40440224]])
        correct_yp = np.array([[-131465.56602927], [-128721.12723285],
                               [-125750.07868119], [-122561.32895953],
                               [-119163.98960288], [-115567.32591061],
                               [-111780.71031054], [-107813.57867853],
                               [-103675.3899576], [-99375.58935766]])

        nptest.assert_almost_equal(xp, correct_xp,
                                   err_msg='Error in tracked x coordinates')

        nptest.assert_almost_equal(yp, correct_yp,
                                   err_msg='Error in tracked y coordinates')

    def test_kick_and_drift_args_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = 10
        machine.values_at_turns()

        denergy = np.array([-115567.32591061])
        dphi = np.array([0.33178332])
        nparts = 1

        recprof = 5
        nturns = machine.dturns * (machine.nprofiles - 1)

        rfv1 = machine.vrf1_at_turn * machine.q
        rfv2 = machine.vrf2_at_turn * machine.q

        xp = np.zeros((machine.nprofiles, nparts))
        yp = np.zeros((machine.nprofiles, nparts))

        phi0 = machine.phi0
        deltaE0 = machine.deltaE0
        omega_rev0 = machine.omega_rev0
        drift_coef = machine.drift_coef
        phi12 = machine.phi12
        h_ratio = machine.h_ratio
        dturns = machine.dturns

        xp, yp = libtomo.kick_and_drift(
            xp, yp, denergy, dphi, rfv1, rfv2, phi0, deltaE0, drift_coef,
            phi12, h_ratio, dturns, recprof, 0, nturns, nparts, False)

        correct_xp = np.array([[0.22739336], [0.24930078], [0.27073013],
                               [0.291644], [0.31200651], [0.33178332],
                               [0.35094168], [0.36945041], [0.38727993],
                               [0.40440224]])
        correct_yp = np.array([[-131465.56602927], [-128721.12723285],
                               [-125750.07868119], [-122561.32895953],
                               [-119163.98960288], [-115567.32591061],
                               [-111780.71031054], [-107813.57867853],
                               [-103675.3899576], [-99375.58935766]])

        nptest.assert_almost_equal(xp, correct_xp,
                                   err_msg='Error in tracked x coordinates')

        nptest.assert_almost_equal(yp, correct_yp,
                                   err_msg='Error in tracked y coordinates')

    def test_kick_and_drift_few_args_fails(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = 10
        machine.values_at_turns()

        denergy = np.array([-115567.32591061])
        dphi = np.array([0.33178332])
        nparts = 1

        recprof = 5
        nturns = machine.dturns * (machine.nprofiles - 1)

        rfv1 = machine.vrf1_at_turn * machine.q
        rfv2 = machine.vrf2_at_turn * machine.q

        xp = np.zeros((machine.nprofiles, nparts))
        yp = np.zeros((machine.nprofiles, nparts))

        phi0 = machine.phi0
        deltaE0 = machine.deltaE0
        omega_rev0 = machine.omega_rev0
        drift_coef = machine.drift_coef
        phi12 = machine.phi12
        h_ratio = machine.h_ratio

        with self.assertRaises(TypeError,
                               msg='Too few arrays should '
                                   'raise an exception'):
            xp, yp = libtomo.kick_and_drift(
                xp, yp, denergy, dphi, rfv1, rfv2, phi0, deltaE0, drift_coef,
                phi12, h_ratio, recprof, nturns, nparts)

        dturns = machine.dturns
        some_useless_var = None
        with self.assertRaises(TypeError,
                               msg='Too many arrays should '
                                   'raise an exception'):
            xp, yp = libtomo.kick_and_drift(
                xp, yp, denergy, dphi, rfv1, rfv2, phi0, deltaE0, drift_coef,
                phi12, h_ratio, some_useless_var, dturns, recprof, nturns,
                nparts)

    def test_back_project(self):
        waterfall = self._load_waterfall()

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        nbins = len(waterfall[0])

        weights = np.zeros(nparts, dtype=np.float64)

        flat_profs = np.ascontiguousarray(
            waterfall.flatten()).astype(np.float64)

        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]
        xp = xp.T
        flat_points = xp.copy()
        for i in range(nprofs):
            flat_points[:, i] += nbins * i
        flat_points = np.ascontiguousarray(flat_points).astype(np.int32)

        libtomo.back_project(weights, flat_points, flat_profs, nparts, nprofs)

        cweights = np.array([0.79054276, 0.80812089, 0.81983964, 0.83468339,
                             0.84327714, 0.84679276, 0.85187089, 0.85187089,
                             0.85148026, 0.85304276, 0.85304276, 0.84601151,
                             0.84444901, 0.84249589, 0.84054276, 0.83937089,
                             0.84405839, 0.83780839, 0.84249589, 0.84796464,
                             0.84483964, 0.85421464, 0.86319901, 0.87530839,
                             0.88038651, 0.88976151, 0.90030839, 0.90733964,
                             0.91671464, 0.92491776, 0.93077714, 0.93937089,
                             0.93780839, 0.94366776, 0.94210526, 0.93780839,
                             0.93312089, 0.93194901, 0.92804276, 0.92530839,
                             0.91749589, 0.91124589, 0.90968339, 0.90421464,
                             0.89718339, 0.89132401, 0.89171464, 0.88780839,
                             0.88273026, 0.88273026])

        nptest.assert_almost_equal(
            weights, cweights, err_msg='Weights were calculated incorrectly')

    def test_project_correct(self):
        nprofs = 10
        nparts = 50
        nbins = 100

        waterfall_shape = (nprofs, nbins)

        weights = np.array([0.79054276, 0.80812089, 0.81983964, 0.83468339,
                            0.84327714, 0.84679276, 0.85187089, 0.85187089,
                            0.85148026, 0.85304276, 0.85304276, 0.84601151,
                            0.84444901, 0.84249589, 0.84054276, 0.83937089,
                            0.84405839, 0.83780839, 0.84249589, 0.84796464,
                            0.84483964, 0.85421464, 0.86319901, 0.87530839,
                            0.88038651, 0.88976151, 0.90030839, 0.90733964,
                            0.91671464, 0.92491776, 0.93077714, 0.93937089,
                            0.93780839, 0.94366776, 0.94210526, 0.93780839,
                            0.93312089, 0.93194901, 0.92804276, 0.92530839,
                            0.91749589, 0.91124589, 0.90968339, 0.90421464,
                            0.89718339, 0.89132401, 0.89171464, 0.88780839,
                            0.88273026, 0.88273026], dtype=np.float64)

        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]
        xp = xp.T
        flat_points = xp.copy()
        for i in range(nprofs):
            flat_points[:, i] += nbins * i
        flat_points = np.ascontiguousarray(flat_points).astype(np.int32)

        rec = libtomo.project(np.zeros(waterfall_shape), flat_points,
                              weights, nparts, nprofs, nbins)

        correct = [0.79054276, 0.80812089, 0.81983964, 0.83468339,
                   0.84327714, 0.84679276, 0.85187089, 0.85187089,
                   0.85148026, 0.85304276, 0.85304276, 0.84601151,
                   0.84444901, 0.84249589, 0.84054276, 0.83937089,
                   0.84405839, 0.83780839, 0.84249589, 0.84796464,
                   0.84483964, 0.85421464, 0.86319901, 0.87530839,
                   0.88038651, 0.88976151, 0.90030839, 0.90733964,
                   0.91671464, 0.92491776, 0.93077714, 0.93937089,
                   0.93780839, 0.94366776, 0.94210526, 0.93780839,
                   0.93312089, 0.93194901, 0.92804276, 0.92530839,
                   0.91749589, 0.91124589, 0.90968339, 0.90421464,
                   0.89718339, 0.89132401, 0.89171464, 0.88780839,
                   0.88273026, 0.88273026, 0., 0., 0., 0., 0., 0.,
                   0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                   0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                   0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                   0., 0., 0., 0., 0., 0., 0., 0.]

        correct = np.array([correct] * 10)

        nptest.assert_almost_equal(
            rec, correct,
            err_msg='Error in recreated profile data after projection')

    def test_reconstruct_correct(self):
        nprofs = 10
        nparts = 50
        nbins = 100
        niter = 1

        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]
        xp = xp.T

        waterfall = self._load_waterfall()
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]

        (weights,
         discr,
         recreated) = libtomo.reconstruct(
            xp, waterfall, niter,
            nbins, nparts, nprofs,
            verbose=False)

        correct_w = np.array([1.40130575, 1.4324645,  1.45323701, 1.47954884,
                              1.49478201, 1.50101376, 1.51001518, 1.51001518,
                              1.50932276, 1.51209243, 1.51209243, 1.49962893,
                              1.49685926, 1.49339718, 1.48993509, 1.48785784,
                              1.49616684, 1.48508818, 1.49339718, 1.50309101,
                              1.49755168, 1.51416968, 1.53009527, 1.55156019,
                              1.5605616,  1.57717961, 1.59587486, 1.60833836,
                              1.62495636, 1.63949711, 1.64988337, 1.66511653,
                              1.66234687, 1.67273312, 1.66996345, 1.66234687,
                              1.65403787, 1.65196062, 1.64503645, 1.64018953,
                              1.6263412,  1.61526253, 1.61249286, 1.60279903,
                              1.59033552, 1.57994927, 1.58064169, 1.57371752,
                              1.5647161,  1.5647161])

        correct_discr = np.array([0.0745208, 0.0745208])

        correct_rec = np.array([0.01797798, 0.01837773, 0.01864423, 0.01898179,
                                0.01917723, 0.01925718, 0.01937266, 0.01937266,
                                0.01936378, 0.01939931, 0.01939931, 0.01923941,
                                0.01920388, 0.01915946, 0.01911504, 0.01908839,
                                0.01919499, 0.01905286, 0.01915946, 0.01928383,
                                0.01921276, 0.01942596, 0.01963028, 0.01990566,
                                0.02002114, 0.02023434, 0.02047419, 0.02063409,
                                0.02084729, 0.02103384, 0.02116709, 0.02136252,
                                0.02132699, 0.02146024, 0.02142471, 0.02132699,
                                0.02122039, 0.02119374, 0.02110491, 0.02104272,
                                0.02086506, 0.02072292, 0.02068739, 0.02056303,
                                0.02040313, 0.02026988, 0.02027876, 0.02018993,
                                0.02007444, 0.02007444, 0., 0., 0., 0., 0., 0.,
                                0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                                0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                                0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                                0., 0., 0., 0., 0., 0., 0., 0.])

        nptest.assert_almost_equal(
            weights, correct_w,
            err_msg='Error in weights after reconstruction.')

        nptest.assert_almost_equal(
            discr, correct_discr,
            err_msg='Error in calculated discrepancies after reconstruction.')

        for i in range(nprofs):
            nptest.assert_almost_equal(recreated[i], correct_rec,
                                       err_msg='Error in '
                                               'reconstructed waterfall.')

    def test_callback(self):
        machine = mch.Machine(**MACHINE_ARGS)

        denergy = np.array([-115567.32591061])
        dphi = np.array([0.33178332])
        nparts = 1

        recprof = 5
        nturns = machine.dturns * (machine.nprofiles - 1)

        rfv1 = machine.vrf1_at_turn * machine.q
        rfv2 = machine.vrf2_at_turn * machine.q

        xp = np.zeros((machine.nprofiles, nparts))
        yp = np.zeros((machine.nprofiles, nparts))

        results = []

        def callback(progress: int, total: int):
            results.append(progress)

        libtomo.kick_and_drift(
            xp, yp, denergy, dphi, rfv1, rfv2, machine, recprof, 0,
            nturns, nparts, callback=callback)

        correct = np.arange(1, nturns+1).tolist()

        self.assertListEqual(results, correct)

    def _load_waterfall(self):
        base_dir = os.path.split(os.path.realpath(__file__))[0]
        base_dir = os.path.split(base_dir)[0]
        data_path = os.path.join(base_dir, 'resources')

        waterfall = np.load(os.path.join(
            data_path, 'waterfall_INDIVShavingC325.npy'))
        return waterfall



"""Unit-tests for the Particles class.

Run as python test_particles_class.py in console or via coverage
"""
from __future__ import annotations
import unittest

from .. import commons
import longitudinal_tomography.tracking.machine as mch
import longitudinal_tomography.tracking.particles as pts
from longitudinal_tomography import exceptions as expt

# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestParticles(unittest.TestCase):

    def test_set_coordinates_dphi_denergy_valid(self):
        dphi = [0.2, 0.4, 0.5]
        denergy = [2142.4, 5432.2, 6732.764]
        input_coords = (dphi, denergy)
        ndims = len(input_coords)

        parts = pts.Particles()
        parts.coordinates_dphi_denergy = input_coords

        for i in range(ndims):
            for read, correct in zip(parts.coordinates_dphi_denergy[i],
                                     input_coords[i]):
                self.assertEqual(read, correct,
                                 msg='Something went wrong when setting '
                                     'coordinates in units of dphi and '
                                     'denergy')

    def test_set_coordinates_dphi_denergy_coord_tuple_no_iter_fails(self):
        input_coords = 1

        parts = pts.Particles()

        with self.assertRaises(
                expt.InvalidParticleError,
                msg='Non iterable input should raise an exception'):
            parts.coordinates_dphi_denergy = input_coords

    def test_set_coordinates_dphi_denergy_coord_one_coord_provided_fails(self):
        dphi = [0.2, 0.4, 0.5]
        input_coords = (dphi, )

        parts = pts.Particles()

        with self.assertRaises(
                expt.InvalidParticleError,
                msg='Providin only one coordinate should raise an exception'):
            parts.coordinates_dphi_denergy = input_coords

    def test_set_coordinates_dphi_denergy_coord_no_iter_fails(self):
        dphi = 0.2
        denergy = 2142.4
        input_coords = (dphi, denergy)

        parts = pts.Particles()

        with self.assertRaises(
                expt.InvalidParticleError,
                msg='Providing non iterable coordinates '
                    'should raise an exception'):
            parts.coordinates_dphi_denergy = input_coords

    def test_set_coordinates_dphi_denergy_coord_bad_len_fails(self):
        dphi = [0.2, 0.4, 0.5]
        denergy = [2142.4, 5432.2]
        input_coords = (dphi, denergy)

        parts = pts.Particles()

        with self.assertRaises(
                expt.InvalidParticleError,
                msg='Providing coordinate arrays of unequal length '
                    'should raise an exception'):
            parts.coordinates_dphi_denergy = input_coords

    def test_homogeneous_distribution_correct_ilimits(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        rbn = 10
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.synch_part_x /= float(rbn)

        parts = pts.Particles()
        parts.homogeneous_distribution(machine, recprof=21)

        imin_correct = 0
        imax_correct = 75

        self.assertEqual(parts.imin, imin_correct,
                         msg='imin has an unexpected value.')
        self.assertEqual(parts.imax, imax_correct,
                         msg='imax has an unexpected value.')

    def test_homogeneous_distribution_correct_jmin(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        rbn = 10
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.synch_part_x /= float(rbn)

        parts = pts.Particles()
        parts.homogeneous_distribution(machine, recprof=21)

        correct = [38, 38, 38, 38, 37, 31, 28, 26, 24, 22, 21, 19, 18, 17,
                   16, 15, 15, 14, 13, 13, 12, 11, 11, 11, 10, 10, 10, 9,
                   9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 11,
                   11, 11, 12, 12, 13, 13, 14, 14, 15, 16, 16, 17, 18, 18,
                   19, 20, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
                   33, 35, 38]

        for j, corr in zip(parts.jmin, correct):
            self.assertEqual(j, corr, msg='Unexpected value found in jmin')

    def test_homogeneous_distribution_correct_jmax(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        rbn = 10
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.synch_part_x /= float(rbn)

        parts = pts.Particles()
        parts.homogeneous_distribution(machine, recprof=21)

        correct = [38, 38, 38, 38, 39, 45, 48, 50, 52, 54, 55, 57, 58, 59,
                   60, 61, 61, 62, 63, 63, 64, 65, 65, 65, 66, 66, 66, 67,
                   67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 66, 66,
                   66, 66, 65, 65, 65, 64, 64, 63, 63, 62, 62, 61, 60, 60,
                   59, 58, 58, 57, 56, 56, 55, 54, 53, 52, 51, 50, 49, 48,
                   47, 46, 45, 43, 41, 38]

        for j, corr in zip(parts.jmax, correct):
            self.assertEqual(j, corr, msg='Unexpected value found in jmin')

    def test_homogeneous_distribution_correct_phase(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        rbn = 250
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.synch_part_x /= float(rbn)

        parts = pts.Particles()
        parts.homogeneous_distribution(machine, recprof=21)

        correct = [-0.24180582, 0.04498875, 0.33178332, 0.61857789,
                   -0.24180582, 0.04498875, 0.33178332, 0.61857789,
                   -0.24180582, 0.04498875, 0.33178332, 0.61857789,
                   -0.24180582, 0.04498875, 0.33178332, 0.61857789]

        for phase, corr in zip(parts.coordinates_dphi_denergy[0], correct):
            self.assertAlmostEqual(
                phase, corr, msg='Error in setting of phase coordinate in '
                                 'cell of phase space at initial distribution')

    def test_homogeneous_distribution_correct_energy(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        rbn = 250
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.synch_part_x /= float(rbn)

        parts = pts.Particles()
        parts.homogeneous_distribution(machine, recprof=21)

        correct = [-115567.32591061, -115567.32591061, -115567.32591061,
                   -115567.32591061, -38522.4419702,   -38522.4419702,
                   -38522.4419702,   -38522.4419702,   38522.4419702,
                   38522.4419702,    38522.4419702,    38522.4419702,
                   115567.32591061,  115567.32591061,  115567.32591061,
                   115567.32591061]

        for energy, corr in zip(parts.coordinates_dphi_denergy[1], correct):
            self.assertAlmostEqual(
                energy, corr, msg='Error in setting of phase coordinate in '
                                  'cell of phase space at initial distribution'
            )



# major_version = 3
# minor_version = 4
# patch_level = 3

# dev_version = -1

# __version__ = '{}.{}'.format(
#     major_version,
#     minor_version,
# )

# if patch_level != 0:
#     __version__ += f'.{patch_level}'

# if dev_version != -1:
#     __version__ += '-dev{}'.format(dev_version)

def use_gpu():
    from .utils import tomo_config as conf
    conf.AppConfig.use_gpu()

def use_cpu():
    from .utils import tomo_config as conf
    conf.AppConfig.use_cpu()

def set_double_precision():
    from .utils import tomo_config as conf
    conf.AppConfig.set_double_precision()

def set_single_precision():
    from .utils import tomo_config as conf
    conf.AppConfig.set_single_precision()

from ._version import version as __version__




"""Module containing ParticleTracker class,
a super class for particle trackers.

:Author(s): **Christoffer Hjertø Grindheim**, **Anton Lu**
"""
from __future__ import annotations
import logging
from typing import TYPE_CHECKING

from . import particles as pts
from .machine_base import MachineABC
from .. import assertions as asrt, exceptions as expt

if TYPE_CHECKING:
    from ..data.profiles import Profiles


log = logging.getLogger(__name__)


class ParticleTracker:
    """Super class for classes meant tracking particles.

    This class holds some general utilities for tracking particles. These
    utilities includes assertions and flags.

    Parameters
    ----------
    machine: MachineABC
        Holds all information needed for particle tracking and generating
        the particle distribution.

    Attributes
    ----------
    machine: MachineABC
        Holds all information needed for particle tracking and generation of
        the particle distribution.
    particles: Particles
        Creates and/or stores the initial distribution of particles.
    nturns: int
        Number of machine turns of which the particles should
        be tracked trough.
    _self_field_flag: boolean
        Flag to indicate that self-fields should be included
        during the tracking.
    _ftn_flag: boolean
        Flag to indicate that the particle tracking should print Fortran-style
        output strings to stdout during tracking.

    Raises
    ------
    MachineParameterError: Exception
        Input argument is not
        :class:`~longitudinal_tomography.tracking.machine.Machine`, or the
        Machine object provided is missing needed fields.
    """

    def __init__(self, machine: MachineABC):

        if not isinstance(machine, MachineABC):
            err_msg = 'Input argument must be Machine.'
            raise expt.MachineParameterError(err_msg)

        self._assert_machine(machine)
        self.machine = machine
        self.particles = pts.Particles()

        self.nturns = machine.dturns * (machine.nprofiles - 1)
        self._self_field_flag = False
        self._ftn_flag = False

        self._profile_charge = None
        self._phiwrap = None
        self._vself = None

    @property
    def self_field_flag(self):
        """self_field_flag defined as @property

        Flag can be set to true by calling :func:`enable_self_fields`.

        Returns
        -------
        self_field_flag: boolean
            Flag to indicate if particle tracking using
            self-fields is enabled.
        """
        return self._self_field_flag

    @property
    def fortran_flag(self):
        """self_field_flag defined as @property

        Flag can be set to true by calling :func:`enable_fortran_output`.

        Returns
        -------
        fortran_flag: boolean
            Flag to indicate if Fortran styled output is enabled.
        """
        return self._ftn_flag

    def enable_fortran_output(self, profile_charge: float):
        """Function for enabling of Fortran-styled output.

        Call this function in order to print a Fortran-styled output
        to stdout during the particle tracking.

        The output will initially be a print of the plot info needed for
        the tomoscope application. Then, During the tracking, the output will
        print which profiles the particles are currently being tracked
        between.

        The number of 'lost particles' will be also be printed.
        This is however **not a real measurement**, but a static string needed
        for the interface to the tomoscope application.
        The lost particles is not found during the tracking due to changes in
        the algorithm.

        Parameters
        ----------
        profile_charge: float
            Total charge of a reference profile.

        Raises
        ------
        ProfileChargeNotCalculated: Exception
            Needed field for enabling Fortran output,
            profile_charge, is missing from the Machine object.
        """
        if profile_charge is None:
            err_msg = 'profile_charge is needed for fortran-style output'
            raise expt.ProfileChargeNotCalculated(err_msg)
        self._ftn_flag = True
        self._profile_charge = profile_charge
        log.info('Fortran style output for particle tracking enabled!')

    def enable_self_fields(self, profiles: 'Profiles'):
        """Function for enabling particle tracking using self-fields.

        Call this function to track the particles using self-fields.
        Note that the self-field tracking is **much slower** than
        tracking without self-fields.

        Parameters
        ----------
        profiles: Profiles
            Self-fields must be calculated in the the Profiles object prior
            to calling this function.
            See :func:`longitudinal_tomography.data.profiles.Profiles.
                        calc_self_fields`.

        Raises
        ------
        SelfFieldTrackingError: Exception
            Needed fields for tracking using self-fields missing\
            from Machine object.
        """
        needed_fieds = ['phiwrap', 'vself']
        asrt.assert_fields(profiles, 'profiles', needed_fieds,
                           expt.SelfFieldTrackingError)

        self._phiwrap = profiles.phiwrap
        self._vself = profiles.vself
        self._self_field_flag = True
        log.info('Tracking using self fields enabled!')

    # Checks that the given machine object includes the necessary
    # variables to perform the tracking.
    # Does not check parameters for calculating using self-fields.
    def _assert_machine(self, machine: MachineABC):
        needed_fieds = ['vrf1_at_turn', 'vrf2_at_turn', 'q',
                        'nprofiles', 'drift_coef', 'dturns', 'phi0',
                        'phi12', 'h_ratio', 'deltaE0', 'synch_part_x']
        asrt.assert_fields(
            machine, 'machine', needed_fieds, expt.MachineParameterError,
            'Did you remember to use machine.values_at_turns()?')
        asrt.assert_greater_or_equal(
            machine.synch_part_x, 'synch_part_x',
            0, expt.MachineParameterError,
            'particle tracking needs a valid synch_part_x value.')



"""Unit-tests for the pre-processing module.

Run as python test_pre_process.py in console or via coverage
"""
from __future__ import annotations
import unittest

import numpy as np
import numpy.testing as nptest

import longitudinal_tomography.data.pre_process as pre_process
import longitudinal_tomography.data.profiles as prf
import longitudinal_tomography.tracking.machine as mch
from .. import commons
import longitudinal_tomography.assertions as asrt
import longitudinal_tomography.exceptions as exceptions

# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestPreProcess(unittest.TestCase):

    def test_rebin_odd_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nbins = 9
        machine.synch_part_x = 4.53
        machine.dtbin = 0.535

        waterfall = np.array([[1, 2, 3, 4, 5, 6, 7, 8, 9],
                              [10, 21, 32, 43, 54, 65, 76, 86, 97]],
                             dtype=float)
        rbn = 2
        (rebinned,
         out_dtbin,
         out_sync_pt_x) = pre_process.rebin(waterfall, rbn, machine.dtbin,
                                            machine.synch_part_x)

        correct_rebinned = np.array([[3.0, 7.0, 11.0, 15.0, 16.0],
                                     [31.0, 75.0, 119.0, 162.0, 172.66666667]])

        nptest.assert_almost_equal(
            rebinned, correct_rebinned,
            err_msg='Rebinning array of odd length failed')

        # Checks that the x coordinate of synchronous particle
        # is updated to fit the new number of bins.
        updated_synch_part_x = 2.265
        self.assertAlmostEqual(out_sync_pt_x, updated_synch_part_x,
                               msg='Error in updated synch part x')

        # Checks that the size of the bins are updated
        # is updated to fit the new number of bins.
        updated_dtbin = 1.07
        self.assertAlmostEqual(out_dtbin, updated_dtbin,
                               msg='Error in updated dtbin')

    def test_rebin_even_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nbins = 8
        machine.synch_part_x = 4.53
        machine.dtbin = 0.535

        waterfall = np.array([[1, 2, 3, 4, 5, 6, 7, 8],
                              [10, 21, 32, 43, 54, 65, 76, 86]],
                             dtype=float)
        rbn = 2
        (rebinned,
         out_dtbin,
         out_sync_pt_x) = pre_process.rebin(waterfall, rbn, machine.dtbin,
                                            machine.synch_part_x)

        correct_rebinned = np.array([[3.0, 7.0, 11.0, 15.0],
                                     [31.0, 75.0, 119.0, 162.0]])

        nptest.assert_almost_equal(
            rebinned, correct_rebinned,
            err_msg='Rebinning array of odd length failed')

        # Checks that the x coordinate of synchronous particle
        # is updated to fit the new number of bins.
        updated_synch_part_x = 2.265
        self.assertAlmostEqual(out_sync_pt_x, updated_synch_part_x,
                               msg='Error in updated synch part x')

        # Checks that the size of the bins are updated
        # is updated to fit the new number of bins.
        updated_dtbin = 1.07
        self.assertAlmostEqual(out_dtbin, updated_dtbin,
                               msg='Error in updated dtbin')

    def test_fit_synch_part_x_correct_x_coord(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.time_at_turn = np.array([1.3701195948153858e-06])
        machine.omega_rev0 = np.array([4585866.32214847])
        machine.phi0 = np.array([0.40078213])

        waterfall = commons.load_waterfall()
        profile = prf.Profiles(machine, machine.dtbin, waterfall)

        correct_x = 134.08972093246575
        fitted_x, _, _ = pre_process.fit_synch_part_x(profile)

        self.assertAlmostEqual(
            fitted_x, correct_x, msg='Fitted x coordinate of synchronous '
                                     'particle is not correct')

        # test function overload
        fitted_x, _, _ = pre_process.fit_synch_part_x(waterfall, machine)
        self.assertAlmostEqual(
            fitted_x, correct_x, msg='Fitted x coordinate of synchronous '
                                     'particle is not correct')

    def test_fit_synch_part_x_correct_tfoot_low(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.time_at_turn = np.array([1.3701195948153858e-06])
        machine.omega_rev0 = np.array([4585866.32214847])
        machine.phi0 = np.array([0.40078213])

        waterfall = commons.load_waterfall()
        profile = prf.Profiles(machine, machine.dtbin, waterfall)

        correct_tfoot_low = 21.569078947368386
        _, tfoot_low, _ = pre_process.fit_synch_part_x(profile)

        self.assertAlmostEqual(
            tfoot_low, correct_tfoot_low,
            msg='Lower tangent foot is calculated incorrectly')

    def test_fit_synch_part_x_correct_tfoot_up(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.time_at_turn = np.array([1.3701195948153858e-06])
        machine.omega_rev0 = np.array([4585866.32214847])
        machine.phi0 = np.array([0.40078213])

        waterfall = commons.load_waterfall()
        profile = prf.Profiles(machine, machine.dtbin, waterfall)

        correct_tfoot_up = 255.84868421052698
        _, _, tfoot_up = pre_process.fit_synch_part_x(profile)

        self.assertAlmostEqual(
            tfoot_up, correct_tfoot_up,
            msg='Upper tangent foot is calculated incorrectly')

    def test_cut_waterfall(self):
        waterfall = np.arange(0, 72).reshape(8, 9)

        correct = np.array([
            [3, 4, 5],
            [12, 13, 14],
            [21, 22, 23],
            [30, 31, 32],
            [39, 40, 41],
            [48, 49, 50],
            [57, 58, 59],
            [66, 67, 68]
        ])

        cut_waterfall = pre_process.cut_waterfall(waterfall, 3, 6)

        np.equal(correct, cut_waterfall)

        cut_waterfall = pre_process.cut_waterfall(waterfall, 3, -3)

        np.equal(correct, cut_waterfall)

    def test_cut_waterfall_bounds(self):
        waterfall = np.arange(0, 72).reshape(8, 9)

        with self.assertRaises(ValueError):
            pre_process.cut_waterfall(waterfall, -1, 0)

        with self.assertRaises(ValueError):
            pre_process.cut_waterfall(waterfall, 1, 10)

        with self.assertRaises(ValueError):
            pre_process.cut_waterfall(waterfall, 3, -6)

        with self.assertRaises(ValueError):
            pre_process.cut_waterfall(waterfall, 3, -10)

        with self.assertRaises(ValueError):
            pre_process.cut_waterfall(waterfall, 5, 5)
